@@ -1,5 +1,5 @@
 /*
-* Copyright (c) 2012-2016 Fredrik Mellbin
+* Copyright (c) 2012-2019 Fredrik Mellbin
 *
 * This file is part of VapourSynth.
 *
@@ -22,6 +22,9 @@
 #include "VSHelper.h"
 #include "filtershared.h"
 #include <stdlib.h>
+#ifdef VS_TARGET_CPU_X86
+#include <emmintrin.h>
+#endif
 
 static inline int CLAMP(int value, int lower, int upper) {
     if (value < lower)
@@ -224,10 +227,6 @@ static void VS_CC preMultiplyCreate(const VSMap *in, VSMap *out, void *userData,
 //////////////////////////////////////////
 // Merge
 
-#ifdef VS_TARGET_CPU_X86
-extern void vs_merge_uint8_sse2(const uint8_t *srcp1, const uint8_t *srcp2, unsigned maskp, uint8_t *dstp, intptr_t stride, intptr_t height);
-#endif
-
 typedef struct {
     VSNodeRef *node1;
     VSNodeRef *node2;
@@ -272,7 +271,30 @@ static const VSFrameRef *VS_CC mergeGetFrame(int n, int activationReason, void *
                     const unsigned round = 1 << (MergeShift - 1);
                     if (d->vi->format->bytesPerSample == 1) {
 #ifdef VS_TARGET_CPU_X86
-                        vs_merge_uint8_sse2(srcp1, srcp2, weight, dstp, stride, h);
+                        __m128i mergeweight = _mm_set1_epi16(weight);
+                        for (int y = 0; y < h; y++) {
+                            for (int xiter = 0; xiter < w; xiter += sizeof(__m128i)) {
+                                __m128i srcreg1 = _mm_load_si128((const __m128i *)(srcp1 + xiter));
+                                __m128i srcreg2 = _mm_load_si128((const __m128i *)(srcp2 + xiter));
+
+                                __m128i srcreg1lo = _mm_unpacklo_epi8(srcreg1, _mm_setzero_si128());
+                                __m128i srcreg2lo = _mm_unpacklo_epi8(srcreg2, _mm_setzero_si128());
+                                __m128i srcreg1hi = _mm_unpackhi_epi8(srcreg1, _mm_setzero_si128());
+                                __m128i srcreg2hi = _mm_unpackhi_epi8(srcreg2, _mm_setzero_si128());
+
+                                __m128i tmp1lo = _mm_slli_epi16(_mm_sub_epi16(srcreg2lo, srcreg1lo), 1);
+                                __m128i tmp1hi = _mm_slli_epi16(_mm_sub_epi16(srcreg2hi, srcreg1hi), 1);
+
+                                __m128i tmp2lo = _mm_add_epi16(_mm_add_epi16(_mm_mulhi_epi16(tmp1lo, mergeweight), _mm_srli_epi16(_mm_mullo_epi16(tmp1lo, mergeweight), 15)), srcreg1lo);
+                                __m128i tmp2hi = _mm_add_epi16(_mm_add_epi16(_mm_mulhi_epi16(tmp1hi, mergeweight), _mm_srli_epi16(_mm_mullo_epi16(tmp1hi, mergeweight), 15)), srcreg1hi);
+
+                                __m128i tmpdst = _mm_packus_epi16(tmp2lo, tmp2hi);
+                                _mm_store_si128((__m128i *)(dstp + xiter), tmpdst);
+                            }
+                            srcp1 += stride;
+                            srcp2 += stride;
+                            dstp += stride;
+                    }
 #else
                         for (int y = 0; y < h; y++) {
                             for (int x = 0; x < w; x++)
@@ -398,10 +420,6 @@ static void VS_CC mergeCreate(const VSMap *in, VSMap *out, void *userData, VSCor
 //////////////////////////////////////////
 // MaskedMerge
 
-#ifdef VS_TARGET_CPU_X86
-extern void vs_masked_merge_uint8_sse2(const uint8_t *srcp1, const uint8_t *srcp2, const uint8_t *maskp, uint8_t *dstp, intptr_t stride, intptr_t height);
-#endif
-
 typedef struct {
     const VSVideoInfo *vi;
     VSNodeRef *node1;
@@ -541,7 +559,35 @@ static const VSFrameRef *VS_CC maskedMergeGetFrame(int n, int activationReason,
                     if (d->vi->format->sampleType == stInteger) {
                         if (d->vi->format->bytesPerSample == 1) {
 #ifdef VS_TARGET_CPU_X86
-                            vs_masked_merge_uint8_sse2(srcp1, srcp2, maskp, dstp, stride, h);
+                            for (int y = 0; y < h; y++) {
+                                for (int xiter = 0; xiter < w; xiter += sizeof(__m128i)) {
+                                    __m128i srcreg1 = _mm_load_si128((const __m128i *)(srcp1 + xiter));
+                                    __m128i srcreg2 = _mm_load_si128((const __m128i *)(srcp2 + xiter));
+                                    __m128i maskreg = _mm_load_si128((const __m128i *)(maskp + xiter));
+
+                                    __m128i srcreg1lo = _mm_unpacklo_epi8(srcreg1, _mm_setzero_si128());
+                                    __m128i srcreg2lo = _mm_unpacklo_epi8(srcreg2, _mm_setzero_si128());
+                                    __m128i maskreglo = _mm_unpacklo_epi8(maskreg, _mm_setzero_si128());
+                                    __m128i srcreg1hi = _mm_unpackhi_epi8(srcreg1, _mm_setzero_si128());
+                                    __m128i srcreg2hi = _mm_unpackhi_epi8(srcreg2, _mm_setzero_si128());
+                                    __m128i maskreghi = _mm_unpackhi_epi8(maskreg, _mm_setzero_si128());
+
+                                    __m128i tmp1lo = _mm_slli_epi16(_mm_sub_epi16(srcreg2lo, srcreg1lo), 4);
+                                    __m128i tmp1hi = _mm_slli_epi16(_mm_sub_epi16(srcreg2hi, srcreg1hi), 4);
+
+                                    __m128i masktmplo = _mm_slli_epi16(_mm_add_epi16(maskreglo, _mm_srli_epi16(_mm_cmpgt_epi16(maskreglo, _mm_set1_epi16(2)), 15)), 4);
+                                    __m128i masktmphi = _mm_slli_epi16(_mm_add_epi16(maskreghi, _mm_srli_epi16(_mm_cmpgt_epi16(maskreghi, _mm_set1_epi16(2)), 15)), 4);
+
+                                    __m128i tmp2lo = _mm_add_epi16(_mm_add_epi16(_mm_mulhi_epi16(tmp1lo, masktmplo), _mm_slli_epi16(_mm_mullo_epi16(tmp1lo, masktmplo), 15)), srcreg1lo);
+                                    __m128i tmp2hi = _mm_add_epi16(_mm_add_epi16(_mm_mulhi_epi16(tmp1hi, masktmphi), _mm_slli_epi16(_mm_mullo_epi16(tmp1hi, masktmphi), 15)), srcreg1hi);
+
+                                    __m128i tmpdst = _mm_packus_epi16(tmp2lo, tmp2hi);
+                                    _mm_store_si128((__m128i *)(dstp + xiter), tmpdst);
+                                }
+                                srcp1 += stride;
+                                srcp2 += stride;
+                                dstp += stride;
+                            }
 #else
                             for (int y = 0; y < h; y++) {
                                 for (int x = 0; x < w; x++)
@@ -560,7 +606,7 @@ static const VSFrameRef *VS_CC maskedMergeGetFrame(int n, int activationReason,
                                     uint16_t s1 = ((const uint16_t *)srcp1)[x];
                                     uint16_t s2 = ((const uint16_t *)srcp2)[x];
                                     uint16_t m = ((const uint16_t *)maskp)[x];
-                                    ((uint16_t *)dstp)[x] = s1 + (((s2 - s1) * (((m >> 1) & 1) + m) + round) >> shift);
+                                    ((uint16_t *)dstp)[x] = s1 + (((int64_t)(s2 - s1) * (((m >> 1) & 1) + m) + round) >> shift);
                                 }
                                 srcp1 += stride;
                                 srcp2 += stride;
@@ -645,7 +691,7 @@ static void VS_CC maskedMergeCreate(const VSMap *in, VSMap *out, void *userData,
     }
 
     if (maskvi->width != d.vi->width || maskvi->height != d.vi->height || maskvi->format->bitsPerSample != d.vi->format->bitsPerSample
-        || (maskvi->format != d.vi->format && maskvi->format->colorFamily != cmGray)) {
+        || (maskvi->format != d.vi->format && maskvi->format->colorFamily != cmGray && !d.first_plane)) {
         vsapi->freeNode(d.node1);
         vsapi->freeNode(d.node2);
         vsapi->freeNode(d.mask);
@@ -681,7 +727,22 @@ static void VS_CC maskedMergeCreate(const VSMap *in, VSMap *out, void *userData,
     // do we need to resample the first mask plane and use it for all the planes?
     if ((d.first_plane && d.vi->format->numPlanes > 1) && (d.vi->format->subSamplingH > 0 || d.vi->format->subSamplingW > 0) && (d.process[1] || d.process[2])) {
         VSMap *min = vsapi->createMap();
-        vsapi->propSetNode(min, "clip", d.mask, paAppend);
+
+        if (maskvi->format->numPlanes > 1) {
+            // Don't resize the unused second and third planes.
+            vsapi->propSetNode(min, "clips", d.mask, paAppend);
+            vsapi->propSetInt(min, "planes", 0, paAppend);
+            vsapi->propSetInt(min, "colorfamily", cmGray, paAppend);
+            VSMap *mout = vsapi->invoke(vsapi->getPluginById("com.vapoursynth.std", core), "ShufflePlanes", min);
+            VSNodeRef *mask_first_plane = vsapi->propGetNode(mout, "clip", 0, 0);
+            vsapi->freeMap(mout);
+            vsapi->clearMap(min);
+            vsapi->propSetNode(min, "clip", mask_first_plane, paAppend);
+            vsapi->freeNode(mask_first_plane);
+        } else {
+            vsapi->propSetNode(min, "clip", d.mask, paAppend);
+        }
+
         vsapi->propSetInt(min, "width", d.vi->width >> d.vi->format->subSamplingW, paAppend);
         vsapi->propSetInt(min, "height", d.vi->height >> d.vi->format->subSamplingH, paAppend);
         VSMap *mout = vsapi->invoke(vsapi->getPluginById("com.vapoursynth.resize", core), "Bilinear", min);
@@ -699,10 +760,6 @@ static void VS_CC maskedMergeCreate(const VSMap *in, VSMap *out, void *userData,
 //////////////////////////////////////////
 // MakeDiff
 
-#ifdef VS_TARGET_CPU_X86
-extern void vs_make_diff_uint8_sse2(const uint8_t *srcp1, const uint8_t *srcp2, uint8_t *dstp, intptr_t stride, intptr_t height);
-#endif
-
 typedef struct {
     VSNodeRef *node1;
     VSNodeRef *node2;
@@ -739,7 +796,18 @@ static const VSFrameRef *VS_CC makeDiffGetFrame(int n, int activationReason, voi
                 if (d->vi->format->sampleType == stInteger) {
                     if (d->vi->format->bytesPerSample == 1) {
 #ifdef VS_TARGET_CPU_X86
-                        vs_make_diff_uint8_sse2(srcp1, srcp2, dstp, stride, h);
+                        __m128i halfpoint = _mm_set1_epi8(0x80);
+                        for (int y = 0; y < h; y++) {
+                            for (int xiter = 0; xiter < w; xiter += sizeof(__m128i)) {
+                                __m128i srcreg1 = _mm_load_si128((const __m128i *)(srcp1 + xiter));
+                                __m128i srcreg2 = _mm_load_si128((const __m128i *)(srcp2 + xiter));
+                                __m128i tmpdst = _mm_add_epi8(_mm_subs_epi8(_mm_sub_epi8(srcreg1, halfpoint), _mm_sub_epi8(srcreg2, halfpoint)), halfpoint);
+                                _mm_store_si128((__m128i *)(dstp + xiter), tmpdst);
+                            }
+                            srcp1 += stride;
+                            srcp2 += stride;
+                            dstp += stride;
+                    }
 #else
                         for (int y = 0; y < h; y++) {
                             for (int x = 0; x < w; x++) {
@@ -855,10 +923,6 @@ static void VS_CC makeDiffCreate(const VSMap *in, VSMap *out, void *userData, VS
 //////////////////////////////////////////
 // MergeDiff
 
-#ifdef VS_TARGET_CPU_X86
-extern void vs_merge_diff_uint8_sse2(const uint8_t *srcp1, const uint8_t *srcp2, uint8_t *dstp, intptr_t stride, intptr_t height);
-#endif
-
 typedef struct {
     VSNodeRef *node1;
     VSNodeRef *node2;
@@ -895,7 +959,18 @@ static const VSFrameRef *VS_CC mergeDiffGetFrame(int n, int activationReason, vo
                 if (d->vi->format->sampleType == stInteger) {
                     if (d->vi->format->bytesPerSample == 1) {
 #ifdef VS_TARGET_CPU_X86
-                        vs_merge_diff_uint8_sse2(srcp1, srcp2, dstp, stride, h);
+                        __m128i halfpoint = _mm_set1_epi8(0x80);
+                        for (int y = 0; y < h; y++) {
+                            for (int xiter = 0; xiter < w; xiter += sizeof(__m128i)) {
+                                __m128i srcreg1 = _mm_load_si128((const __m128i *)(srcp1 + xiter));
+                                __m128i srcreg2 = _mm_load_si128((const __m128i *)(srcp2 + xiter));
+                                __m128i tmpdst = _mm_add_epi8(_mm_adds_epi8(_mm_sub_epi8(srcreg1, halfpoint), _mm_sub_epi8(srcreg2, halfpoint)), halfpoint);
+                                _mm_store_si128((__m128i *)(dstp + xiter), tmpdst);
+                            }
+                            srcp1 += stride;
+                            srcp2 += stride;
+                            dstp += stride;
+                        }
 #else
                         for (int y = 0; y < h; y++) {
                             for (int x = 0; x < w; x++) {
