@@ -29,6 +29,8 @@
 #define vsFatal(...)    vsLog(__FILE__, __LINE__, mtFatal, __VA_ARGS__)
 
 void vsSetMessageHandler(VSMessageHandler handler, void *userData);
+int vsAddMessageHandler(VSMessageHandler handler, VSMessageHandlerFree free, void *userData);
+int vsRemoveMessageHandler(int id);
 void vsLog(const char *file, long line, VSMessageType type, const char *msg, ...);
 
 #endif
